#include "adc.h"
#include "humidity.h"

#define VREF_HI 6.25
#define VREF_LO 1.25

#define UprobeConnected ((P5IN & Px4) == 0)     // If pin 5.4 is low, the probe is connected
#define TPenableInt  (P1SEL |=  Px2)            // enable trip-point interrupt
#define TPdisableInt (P1SEL &= ~Px2)            // disable trip-point interrupt
#define URenableInt (P1IE |= Px1)               // enable under-range interrupt
#define URdisableInt (P1IE &= ~Px1)             // disable under-range interrupt

RampCtrlStruct rctrl[MAX_RAMPS];
int event;                                      // ramp converter event counter, 0,1,2,0,1,2,...
uint muxInput;                                  // input to the multiplexor

extern uint vTimer;                             // main.c

// *********************************************************************
// allowable addresses for the mux are 0 to 15
// every address line is inverted, so invert the address
// *********************************************************************
void SetMuxAddr(uchar addr) {
        TPdisableInt;                           // disable trip-point interrupt
        P3OUT = ((P3OUT & 0xF0) | ((~addr)&0x0F));      // set address lines (inverted)
        TPenableInt;                            // enable trip-point interrupt
}

// *********************************************************************
// start the ramp, tell the ramp to go high
// *********************************************************************
void StartRamp (void) {
	SetMuxAddr(VREF_LO_ADDR);                   // set the mux to look at vref low
	event = 0;                                  // update event state
	P6OUT &= ~Px3;                              // start ramp
}

// *********************************************************************
// stop the ramp, tell the ramp to go low
// *********************************************************************
void StopRamp (void) {
    TPdisableInt;                               // disable trip-point interrupt
	event = -1;                                 // update event state
	P6OUT |= Px3;                               // reset (stop) ramp
}

// *********************************************************************
// compute resistance, given a ramp ratio and the values of the wing resistors around it
// *********************************************************************
float CalcRampRes (float ratio, float lo, float hi) {
    return (hi/ratio - hi - lo);
}

// *********************************************************************
// compute voltage, given a ramp ratio
// *********************************************************************
float CalcRampVolt (float ratio) {
    return (VREF_HI - VREF_LO) * ratio + VREF_LO;
}

// *********************************************************************
// initialize ramp
// *********************************************************************
void RampInit (void) {
    int     i;

    for (i=0;i<MAX_RAMPS;i++) {             // init ramp control data
        rctrl[i].rt.valid = false;
    }
    rctrl[BAT_ID].mux_addr = BAT_ADDR;       // init mux data
    rctrl[U_TEMP_ID].mux_addr = U_TEMP_ADDR;
    rctrl[T1_ID].mux_addr = T1_ADDR;
    rctrl[P_ID].mux_addr = P_ADDR;
    rctrl[P_TEMP_ID].mux_addr = P_TEMP_ADDR;
    rctrl[INT_TEMP_ID].mux_addr = SPARE_11_ADDR;
    P3DIR |= ( Px0 + Px1 + Px2 + Px3 );     // set mux address select pins as outputs
    P6DIR |= Px3;                           // set the ramp converter reset pin as output
    StopRamp();
    muxInput = 0;                           // Mux input control, starts with input 1.
    P3OUT |= 8;                             // this enables mux 1, disables mux 2
    TPenableInt;                            // enable ramp trip point interrupt
    URenableInt;                            // enable under-range interrupt
}

// *********************************************************************
// calc ramp data ratios for all channels, this must happen when the ramp is stopped,
// right before calling "RestartRamp" is a good place to do it.
// *********************************************************************
void RampCalcAllRatio (RampCtrlStruct rcs[]) {
    uchar   i;

    for (i=0;i<MAX_RAMPS;i++) {             // for all the inputs calc ratios
        rcs[i].ratio = RampCalcRatio(&(rcs[i].rt));
    }
}

// *********************************************************************
// calc ramp data ratios 
// *********************************************************************
float RampCalcRatio (RampTimesStruct* rt) {
    float   r;

    if (!rt->valid) {                       // validate data first
        r = -1.0;                           // data not valid, indicate error
    } else {                                // data valid
        if ((rt->end - rt->start) > 0) {    // prevent div by 0
            r = (float)(rt->meas - rt->start) / (float)(rt->end - rt->start);
        } else {                            // total time = 0 (invalid)
            r = -2.0;                       // inidicate div by 0 error
        }
    }
    rt->valid = false;                      // clear flag
    return r;                               // return ratio
}

// *********************************************************************
// start a measurement sequence
// *********************************************************************
void RestartRamp(void) {
    uchar   i;

    muxInput = 0;                           // start with mux pos 0
    for (i=0;i<MAX_RAMPS;i++) {             // init all channels
        rctrl[i].rt.valid = false;           // init flags, nothing measured yet
    }
    if ((P1IN & Px1) == Px1) {              // check if the ramp is off by looking at under-range
        StartRamp();                        // ramp ready to start
    } else {                                // ramp not ready
        StopRamp();                         // reset the ramp to make ready for measurement
    }
}

// *********************************************************************
// ramp trigger interrupt
// *********************************************************************
void RampTrigInt (ulong eventTime) {
    if (event >= 0 && event <= 2) {                         // validate event
        switch (event) {
            case 0:                                         // VREF_LO measured
                rctrl[muxInput].rt.start = eventTime;       // save ramp time
                SetMuxAddr(rctrl[muxInput].mux_addr);       // measure sensor next
                event++;                                    // update event state
                break;
            case 1:                                         // sensor measured
                if ((eventTime - rctrl[muxInput].rt.start) > HYSTERESIS_COUNT) {
                    rctrl[muxInput].rt.meas = eventTime;    // save ramp time
                    SetMuxAddr(VREF_HI_ADDR);               // measure VREF_HI next
                    event++;                                // update event state
                }
                break;
            case 2:                                         // measurement cycle done
                if ((eventTime - rctrl[muxInput].rt.meas) > HYSTERESIS_COUNT) {
                    rctrl[muxInput].rt.end = eventTime;     // save ramp time
                    rctrl[muxInput].rt.valid = true;        // update flag
                    muxInput++;                             // advance to next mux input
                    // do not measure temp or humidity if there is no probe
                    if (!UprobeConnected) {
                        while ((rctrl[muxInput].mux_addr == T1_ADDR) || (rctrl[muxInput].mux_addr == U_TEMP_ADDR)) {
                            muxInput++;                     // skip channels
                        }
                    }
                    StopRamp();                             // shut off the ramp, so it can restart if more channels to measure
                }
                break;
        }
    }
}

// *********************************************************************
// called when the ramp under-range interrupt occurs
// only restart the ramp if the ramp is currently off (event = -1),
// and we still have voltages to measure.
// *********************************************************************
void RampUnderRangeInt (void) {
	if ((event == -1) && (muxInput < MAX_RAMPS)) {      // waiting to start measurement cycle
        StartRamp();                                    // start ramp
    }
}
