#ifndef __adc_h__
#define __adc_h__

/*
    The ramp converter requires three events to be timed for every data 
    conversion. Thus for each conversion the following sequence is followed:
    1)The low reference voltage is selected and the crossing event timed.
    2)The desired data line is selected and the crossing event timed.
    3)The high reference voltage is selected and the corssing event timed.
    Once these three event times are saved the ratio is calculated. Three 
    control lines are programed to select the references and data, this assumes 
    that the hardware is an eight channel MUX. Thus we can concievably convert 
    six data lines.
    
    This program uses Timer_A in a continuous mode with external event
    on P1.2. The event time is captured on the leading (rising) edge of the 
    event and stored. The process is started with the first event time stored.  
    A RAM register is used to extend the timer register to 32 bits. The second 
    and third events are saved and then the ratio of the time from the first 
    event to the second divided by the time from the first to the last is 
    computed. Timer_A register 1 is used for the ramp converter.
*/

#include "globals.h"

#define MAX_RAMPS   6               // maximum number of inputs

#define PT_WING_R1      0.0         // wing resister values for the pressure sensor's temperature thermistor
#define PT_WING_R2      6.04e3
#define IT_WING_R1      0.0         // wing resistor values for the internal tmeperature thermistor
#define IT_WING_R2      6.65e3
#define T_WING_R1       0.0         // wing resister values for the ambient temperature thermistor
#define T_WING_R2       65.0e3
#define UT_WING_R1      0.0         // wing resister values for the humidity sensor's temperature thermistor
#define UT_WING_R2      53.6e3              

#define BAT_ID      0               // battery
#define U_TEMP_ID   1               // humidity temperature
#define INT_TEMP_ID 2               // internal temperature
#define T1_ID       3               // T1 temperature
#define P_ID        4               // pressure
#define P_TEMP_ID   5               // pressure temperature

// allowable addresses for the mux are 0 to 15
// every address line is inverted, so invert the address
#define VREF_HI_ADDR    0
#define VREF_LO_ADDR    1
#define P_ADDR          2
#define P_TEMP_ADDR     3
#define T1_ADDR         4
#define U_TEMP_ADDR     5
#define BAT_ADDR        6
#define SPARE_11_ADDR   7
#define T2_ADDR         8
#define T3_ADDR         9
#define SPARE_21_ADDR   10
#define SPARE_22_ADDR   11
#define SPARE_23_ADDR   12
#define SPARE_24_ADDR   13
#define SPARE_25_ADDR   14
#define SPARE_26_ADDR   15

// make the ramp hysteresis = 1 mS
#define HYSTERESIS_COUNT 8000

typedef struct {
    ulong   start,              // time of first event, vref low
            meas,               // time of second event, sensor
            end;                // time of third event, vref high
    bool    valid;
} RampTimesStruct;

typedef struct {
    uchar   mux_addr;
    RampTimesStruct rt;
    float   ratio;
} RampCtrlStruct;

void RampInit (void);                           // Initialize the ramp, when system boots
void RestartRamp (void);                        // restart the ramp conversion sequence
void RampCalcAllRatio (RampCtrlStruct rcs[]);           // calc ratios for all ramp data
float CalcRampRes (float ratio, float lo, float hi);    // compute resistance, given a ramp ratio and the values of the wing resistors around it
float CalcRampVolt (float ratio);               // compute voltage given ratio
void RampTrigInt (ulong time);                  // call this when the ramp is triggered
void RampUnderRangeInt (void);                  // under-range, continue
void SetMuxAddr (uchar);                        // set the mux to the specified address
void StartRamp (void);                          // start the ramp, tell the ramp to go high
void StopRamp (void);                           // stop the ramp, tell the ramp to go low
float RampCalcRatio (RampTimesStruct* rt);      // perform a single ramp conversion

    
#endif
