#include "adc12.h"
#include "clock.h"
#include "global.h"
#include "config.h"
#include <math.h>

#define ADC12_iTCASE        0               // ADC0
#define ADC12_iTAIR         1               // ADC1
#define ADC12_iFLOW         2               // ADC2

ulong   adcSum[ADC12_NCH];                  // ADC summation, summation of each channel for averaging
float   adcVal[ADC12_NCH];                  // average voltage (previous interval)
ulong   adcCnt[ADC12_NCH];                  // number of samples in summation data

extern CfgStruct        cfg;                // from main.c
extern volatile ulong   flags;              // from main.c

// initialize the ADC
void ADC12Init (void) {
    int     i;

    ADC12CTL0 = SHT00 + SHT01 +             // sampling period = ACLK/32 = 977 uS
                MSC +                       // multiple sample & conversion
                REF2_5V +                   // use internal 2.5v reference, also used by DAC12
                REFON +                     // use internal reference, also used by DAC12
                ADC12ON;                    // enable ADC, conversions not enabled yet
    ADC12CTL1 = 0 +                         // conversion start addres = 0
                SHP +                       // SAMPCON signal is sourced from the sampling timer
                ADC12SSEL0 +                // use ACLK for ADC12CLK
                CONSEQ0 + CONSEQ1;          // Repeat-sequence-of-channels
                                            // ADC12 clock divider = 1
    ADC12MCTL0 = SREF0;                     // internal reference, channel ADC0
    ADC12MCTL1 = SREF0 + INCH0;             // internal reference, channel ADC1
    ADC12MCTL2 = SREF0 + INCH1 +            // internal reference, channel ADC2
                 EOS;                       // end of sequence
    for (i=0;i<ADC12_NCH;i++) {             // init data
        adcVal[i] = 0.0;
        adcSum[i] = 0;
        adcCnt[i] = 0;
    }
    ADC12IE = (1 << ADC12_NCH) - 1;         // enable conversion complete interrupts
    ADC12CTL0 |= ENC + ADC12SC;             // start conversions
}


void ADC12 (void) __interrupt[ADC12_VECTOR] {       // ADC interrupt service routine
    uint    i;

     switch(ADC12IV) {
        case ADC12IV_ADC12OVIFG:
            ADC12IV &= ~ADC12IV_ADC12OVIFG;
            break;
        case ADC12IV_ADC12TOVIFG:
            ADC12IV &= ~ADC12IV_ADC12TOVIFG;
            break;
        case ADC12IV_ADC12IFG0:
            adcSum[0] += ADC12MEM0;                 // get data from all channels (Repeat-sequence-of-channels mode)
            adcCnt[0]++;
            adcSum[1] += ADC12MEM1;
            adcCnt[1]++;
            adcSum[2] += ADC12MEM2;
            adcCnt[2]++;
            break;
        case ADC12IV_ADC12IFG1:
            i = ADC12MEM1;
            break;
        case ADC12IV_ADC12IFG2:
            i = ADC12MEM2;
            break;
        case ADC12IV_ADC12IFG3:
            i = ADC12MEM3;
            break;
        case ADC12IV_ADC12IFG4:
            i = ADC12MEM4;
            break;
        case ADC12IV_ADC12IFG5:
            i = ADC12MEM5;
            break;
        case ADC12IV_ADC12IFG6:
            i = ADC12MEM6;
            break;
        case ADC12IV_ADC12IFG7:
            i = ADC12MEM7;
            break;
        case ADC12IV_ADC12IFG8:
            i = ADC12MEM8;
            break;
        case ADC12IV_ADC12IFG9:
            i = ADC12MEM9;
            break;
        case ADC12IV_ADC12IFG10:
            i = ADC12MEM10;
            break;
        case ADC12IV_ADC12IFG11:
            i = ADC12MEM11;
            break;
        case ADC12IV_ADC12IFG12:
            i = ADC12MEM12;
            break;
        case ADC12IV_ADC12IFG13:
            i = ADC12MEM13;
            break;
        case ADC12IV_ADC12IFG14:
            i = ADC12MEM14;
            break;
        case ADC12IV_ADC12IFG15:
            i = ADC12MEM15;
            break;
    }
}


void ADC12avg (void) __monitor {
    int     i;

    for (i=0;i<ADC12_NCH;i++) {
        if (adcCnt[i] > 0) {                    // prevent div by 0
            adcVal[i] = ((float)adcSum[i] / (float)adcCnt[i]) / (float)ADC12_FULL_SCALE;      // calc average 0.0 - 1.0 (FS)
        } else {
            adcVal[i] = 0.0;
        }
        adcSum[i] = 0;                          // reset summation
        adcCnt[i] = 0;
    }
}


// return current channel voltage
float ADC12ChV (uchar ch) __monitor {

    if (ch < ADC12_NCH) {                                       // validate requested channel
        return adcVal[ch] * (float)ADC12_VREF;                  // return voltage
    } else {                                                    // invalid channel
        return 0.0;
    }
}


// case temperature (C)
float Tcase (void) __monitor {
    float       v;

    flags &= ~F_TEMP_ERR;                                       // clear error flag
    if (adcVal[ADC12_iTCASE] < ADC12_VREF) {                                            // prevent div by 0
        v = (adcVal[ADC12_iTCASE] * cfg.Tcase.Rref) / (1 - adcVal[ADC12_iTCASE]);       // calculate resistance
        v = log(v);
        v = ((cfg.Tcase.K[3] * v + cfg.Tcase.K[2]) * v + cfg.Tcase.K[1]) * v + cfg.Tcase.K[0];
        if (v != 0.0) {
            v = (1.0 / v) - 273.15;                             // temperature in deg C
        } else {
            v = TCASE_MAX;                                      // assign max value
            flags |= F_TEMP_ERR;                                // set error flag
        }
    } else {
        v = TCASE_MAX;                                          // assign max value
        flags |= F_TEMP_ERR;                                    // set error flag
    }
    if (v > TCASE_MAX) {                                        // check high limit
        v = TCASE_MAX;
        flags |= F_TEMP_ERR;                                    // set error flag
    }
    if (v < TCASE_MIN) {                                        // check low limit
        v = TCASE_MIN;
        flags |= F_TEMP_ERR;                                    // set error flag
    }
    return v;
}


// air temperature (C)
float Tair (void) __monitor {
    float       v;

    flags &= ~F_TEMP_ERR;                                       // clear error flag
    if (adcVal[ADC12_iTAIR] < ADC12_VREF) {                                             // prevent div by 0
        v = (adcVal[ADC12_iTAIR] * cfg.Tair.Rref) / (1 - adcVal[ADC12_iTAIR]);          // calculate resistance
        v = log(v);
        v = ((cfg.Tair.K[3] * v + cfg.Tair.K[2]) * v + cfg.Tair.K[1]) * v + cfg.Tair.K[0];
        if (v != 0.0) {
            v = (1.0 / v) - 273.15;                             // temperature in deg C
        } else {
            v = TAIR_MAX;                                       // assign max value
            flags |= F_TEMP_ERR;                                // set error flag
        }
    } else {
        v = TAIR_MAX;                                           // assign max value
        flags |= F_TEMP_ERR;                                    // set error flag
    }
    if (v > TAIR_MAX) {                                         // check high limit
        v = TAIR_MAX;
        flags |= F_TEMP_ERR;                                    // set error flag
    }
    if (v < TAIR_MIN) {                                         // check low limit
        v = TAIR_MIN;
        flags |= F_TEMP_ERR;                                    // set error flag
    }
    return v;
}


// std flow rate (m/s)
float Flow (void) __monitor {
    float       v;

    flags &= ~F_FLOW_ERR;                                       // clear error flag
    v = ADC12ChV(ADC12_iFLOW)*4.0;                              // calc sensor output voltage
    v = ((cfg.Flow.K[3] * v + cfg.Flow.K[2]) * v + cfg.Flow.K[1]) * v + cfg.Flow.K[0];
    if (v < FLOW_MIN) {                                         // check data
        flags |= F_FLOW_ERR;                                    // set error flag
        v = FLOW_MIN;
    }
    if (v > FLOW_MAX) {                                         // check data
        flags |= F_FLOW_ERR;                                    // set error flag
        v = FLOW_MAX;
    }
    return v;
}
