#include <string.h>
#include "clock.h"
#include "global.h"
#include "led.h"
#include "adc12.h"
#include "config.h"
#include "xadc.h"

extern  volatile ulong  flags;              // from main.c
extern  float           ledPWM[];           // from main.c
extern  CfgStruct       cfg;                // from main.c
extern  ulong           elapsedTime;        // from main.c

uint        outCnt;                         // 1-sec output counter


// initialize main clock sources
// MCLK & SMCLK = 10.00MHz, ACLK = 32.768KHz
void ClockInit(void) {
    int     i;

    DCOCTL = DCO2+DCO1;                     // setup DCO = 6.00-9.60 MHz
    BCSCTL1 = RSEL3 + RSEL2 + RSEL0;        // XT2 = on (10.00 MHz),
                                            // LFXT1 = LF XTAL (32.768 KHz),
    BCSCTL3 = XT2S1;                        // 3 - 16-MHz crystal
                                            // 32768 Hz Crystal on LFXT1
    do {                                    // start clocks
        IFG1 &= ~OFIFG;                     // clear OSCFault flag
        for (i=0; i<100; i++) {};           // wait 100uS (at least 50 uS required)
    } while ((IFG1 & OFIFG) != 0);          // OSCFault flag still set?
    BCSCTL2 = SELM1 + SELS;                 // MCLK = SMCLK = XT2
    elapsedTime = 0;                        // start time
}


// initialize Timer A
void TimerA_Init(void) {
    outCnt = 0;                             // 1-sec interval counter
    TACTL = TASSEL1 +                       // Timer A clock = SMCLK (10 MHz)
            ID0 +                           // input divider = SMCLK / 2
            MC0 +                           // mode = up
            TACLR;                          // clear clk
                                            // overflow interrupt set, counter cleared

    TACCR0 = TA_INTERVAL-1;                 // set counter max
    TACCR1 = TA_INTERVAL / 2;               // default LED intensity PWM (50%)
    TACCR2 = 0;
    TACCTL0 = 0;
    TACCTL1 = OUTMOD2+OUTMOD1+OUTMOD0;      // reset-set mode, no interrupt
    TACCTL2 = OUTMOD2;                      // toggle mode, no interrupt
}


// initialize Timer B
// 16-bit, independent latches
void TimerB_Init(void) {
    TBCTL = TBSSEL0+MC0+TBIE+TBCLR;         // Timer A clock = ACLK (32.768 KHz), up mode,
                                            // overflow interrupt set, counter cleared

    TBCCR0 = TB_INTERVAL-1;                 // 1-sec interval
    TBCCR1 = 0;                             // heater 1 (case), off at startup
    TBCCR2 = 0;                             // heater 2 (aux), off at startup
    TBCCR3 = (TB_INTERVAL/4)-1;             // red LED (1/4 of interval)
    TBCCR4 = (TB_INTERVAL/4)*2-1;           // green LED (1/4 of interval)
    TBCCR5 = (TB_INTERVAL/4)*3-1;           // blue LED (1/4 of interval)
    TBCCR6 = 0;
    TBCCTL0 = 0;
    TBCCTL1 = OUTMOD2+OUTMOD1;              // set/reset mode
    TBCCTL2 = OUTMOD2+OUTMOD1;              // set/reset mode
    TBCCTL3 = CCIE;                         // manual output mode (red LED), enable interrupt
    TBCCTL4 = CCIE;                         // manual output mode (green LED), enable interrupt
    TBCCTL5 = CCIE;                         // manual output mode (blue LED), enable interrupt
    TBCCTL6 = 0;
}


// Timer B interrupt handler
// CC1-6, TB
void TimerB1(void) __interrupt[TIMERB1_VECTOR] {
    if (!(FPSW_PIN & FPSW)) {                           // check for button press, 0=pressed
        flags |= F_FPSW_PRESS;                          // set flag
    }
    switch (TBIV) {
        case TBIV_TBIFG:                                // TB overflow
            SetAllLEDs(NONE,RED+GREEN+BLUE);            // all off (global flags updated)
            flags |= F_DATA_UPDATE;                     // time to average 1-sec data
            ADC12avg();                                 // calc average ADC12 data over 1-sec interval
            outCnt = ++outCnt % cfg.outputInterval;     // update loop counter
            if (outCnt == 0) {                          // time to ouput data
                flags |= F_DATA_OUT;                    // time to send data
            }
            elapsedTime++;                              // update time
            break;
        case TBIV_TBCCR1:                               // heater 1 (case)
            break;
        case TBIV_TBCCR2:                               // heater 2 (aux)
            break;
        case TBIV_TBCCR3:                               // red LED gate
            SetAllLEDs(RED,GREEN+BLUE);                 // red on, all others off (global flags updated)
            Set_LEDintensity(cfg.red);                  // set brightness
            break;
        case TBIV_TBCCR4:                               // green LED gate
            SetAllLEDs(GREEN,RED+BLUE);                 // green on, all others off (global flags updated)
            Set_LEDintensity(cfg.grn);                  // set brightness
            break;
        case TBIV_TBCCR5:                               // blue LED gate
            SetAllLEDs(BLUE,RED+GREEN);                 // blue on, all others off (global flags updated)
            Set_LEDintensity(cfg.blue);                 // set brightness
            break;
    }
}

