#include "configuration.h"
#include "crc.h"
#include "flash.h"
#include "humidity.h"
#include "pressure.h"
#include "temperature.h"
#include "transmitter.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>

CfgParamStruct cfg;         // configuration parameters

// ****************************************************************************
// read configuration structure from FLASH
// return 0 if ok, anything else is an error
// ****************************************************************************
uchar RdCfgFlash (void) {
    uint    crc;
    int     i;

    if (RdInfoFlash((char*)&cfg,sizeof(cfg))) {                     // error reading config from FLASH
        return 1;
    }
    crc = calc_crc((char*)&cfg,0,sizeof(cfg),0);        			// calc CRC
    if (crc) {                                                      // error in cfg
        return 2;
    }
    return 0;               // read ok
}

// ****************************************************************************
// write configuration structure to FLASH
// return 0 if ok, anything else is an error
// ****************************************************************************
uchar WrCfgFlash (void) {
    int     i;

    update_crc((char*)&cfg,0,sizeof(cfg),0);      					// update crc in struct
    if (WrInfoFlash((char*)&cfg,sizeof(cfg))) {                     // error writing to FLASH
        return 1;
    }
    return 0;               // write ok
}

// ****************************************************************************
// init configuration data
// ****************************************************************************
void InitCfg (void) {
    memset(&cfg, 0, sizeof(cfg));               // clear struct
    // set some reasonable defaults...
    InitTaCoeff(cfg.cal_coeff.Ta);              // ambient temperature, shibaura thermistor
    InitPCoeff(cfg.cal_coeff.P);                // pressure coefficients
    InitTpCoeff(&cfg.cal_coeff.Tpr);            // pressure temperature
    InitUCoeff(cfg.cal_coeff.U);                // humidity
    InitTuCoeff(cfg.cal_coeff.Tu);              // temperature of humidity sensor, humirel thermistor
    memcpy(cfg.sn, "000000000\0", 10);
    memcpy(cfg.probe, "P000000\0", 8);
    memcpy(cfg.sonde, "S000000\0", 8);
    if (TX_IS_403) {                            // 403MHz transmitter default frequencies
        cfg.tx_freq[0] = 402.0;
        cfg.tx_freq[1] = 403.0;
        cfg.tx_freq[2] = 404.0;
        cfg.tx_freq[3] = 405.0;
    } else {                                    // 1680MHz transmitter default frequencies
        cfg.tx_freq[0] = 1676.0;
        cfg.tx_freq[1] = 1678.0;
        cfg.tx_freq[2] = 1680.0;
        cfg.tx_freq[3] = 1682.0;
    }
    cfg.tx_mode = BEL_202;                      // start with BEL202, 1200 baud
    cfg.heater_sp = HEATER_SETPT;               // heater turn on temperature
    cfg.flags = F_TXRF;                         // transmitter on
    cfg.flags |= F_LEDS;                        // LEDs on
    cfg.flags |= F_PTUX;                        // enhanced PTU data on
    cfg.rrPTU = RS_REPORT_RATE;                 // default report rates
    cfg.rrGPS = GPS_REPORT_RATE;
    cfg.Ucrc = 0;                               // init U-table CRC
}

// ****************************************************************************
// read switch position and set transmitter frequency
// ****************************************************************************
float TxFreq (void) {
    float freq;

    freq = cfg.tx_freq[GetSwPos()];
    SetTXfreq(freq);
    return freq;
}

// ****************************************************************************
// read switch position
// ****************************************************************************
uchar GetSwPos (void) {
    uchar value, pos;

    value = (P4IN & (Px3 | Px4 | Px5))/8;       // read pins
    pos = 0;
    switch(value) {             // translate to position
        case 3: pos=0;
                break;
        case 5: pos=1;
                break;
        case 7: pos=2;
                break;
        case 6: pos=3;
                break;
    }
    return pos;
}



