#include "flash.h"
#include "global.h"

#include <string.h>

void EraseSegment(ulong segment);
void WaitFlash(void);
int WrFlashSegment(char* data, ulong segment, uint size);


// read data from information flash
int RdInfoFlash(char* data, uint size) {
    if (size <= INFO_SEGMENT_SIZE * INFO_SEGMENT_CNT) {     // validate data size
        memcpy(data,(char*)SEGMENT_D, size);                // read data from flash
        return 0;
    } else {                                                // error, size too large
        return 1;
    }
}


// read data from any segment in flash
int RdFlash(char* data, ulong segment, uint size) {
    int     i;

    if ((uint)segment + size - 1 <= TOP_OF_FLASH) {         // validate file size
        // memcpy(data,(void*)segment,size);                          // read data from flash
        for (i=0;i<size;i++) {
            data[i] = __read_extended_byte(segment+i);
        }
        return 0;
    } else {                                                // error size too large
        return 1;
    }
}


// wait for flash to be ready (blocking code)
void WaitFlash(void) {
    while ((FCTL3 & BUSY)|(~FCTL3 & WAIT)) {};
}


// erase segment of flash
void EraseSegment(ulong segment)
{
    FCTL1 = FWKEY + ERASE;                      // erase individual segment only
    FCTL3 = FWKEY;                              // unlock flash memory for write/erase
    __write_extended_byte(segment,0);           // dummy write to anywhere in flash segment forces erase
}


int WrFlashSegment (char* data, ulong segment, uint size) {
    int     i;
    uint    sz;

    if ((segment == SEGMENT_A) || (segment == SEGMENT_B) ||
        (segment == SEGMENT_C) || (segment == SEGMENT_D)) {
        sz = INFO_SEGMENT_SIZE;
    } else {
        sz = SEGMENT_SIZE;
    }
    if (size > sz) {                            // error, size too big
        return 1;                               // error code
    }
    WaitFlash();
    FCTL2 = FWKEY + FSSEL1 + FN4+FN3;           // set flash clock = smclk/0x18 = 10 MHz/(FNx+1) = 400kHz
    FCTL3 = FWKEY+LOCKA;                        // unlock flash for writing/erasing (including SEGMENT_A)
    EraseSegment(segment);                      // erase segment
    FCTL1 = FWKEY + WRT;                        // turn on write mode
    for (i=0; i<size; i++) {
        WaitFlash();
//        segment[i] = data[i];                   // write data
        __write_extended_byte(segment+i, data[i]);
    }
    FCTL1 = FWKEY;                              // turn off write mode
    FCTL3 = FWKEY + LOCK;                       // lock flash, writing/erasing not allowed
    BCSCTL2 = SELM1 | SELS;                     // Set the clock back to 8MHz
    return 0;
}


// write data to information flash
int WrInfoFlash(char* data, uint size) __monitor {
    uint    n,sz;

    if (size > INFO_SEGMENT_SIZE * INFO_SEGMENT_CNT) {      // error, size too big
        return 1;                                           // return error code
    }
    if (size >= INFO_SEGMENT_SIZE) {                        // calc size for writing info_segment
        sz = INFO_SEGMENT_SIZE;
    } else {
        sz = size;
    }
    if (WrFlashSegment(data, (ulong)SEGMENT_D, sz)) {       // validate write
        return 2;                                           // error, return code
    }
    if (size > INFO_SEGMENT_SIZE) {                                     // data exists to write
        n = (uint)data + INFO_SEGMENT_SIZE;                             // update pointer for next segment write
        if (size-INFO_SEGMENT_SIZE >= INFO_SEGMENT_SIZE) {              // calc size for writing info_segment
            sz = INFO_SEGMENT_SIZE;
        } else {
            sz = size-INFO_SEGMENT_SIZE;
        }
        if (WrFlashSegment((char*)n, (ulong)SEGMENT_C, sz)) {           // write next segment
            return 3;                                                   // error, return code
        }
        if (size > INFO_SEGMENT_SIZE*2) {                               // data exists to write
            n = n + INFO_SEGMENT_SIZE;                                  // update pointer for next segment write
            if (size-INFO_SEGMENT_SIZE*2 >= INFO_SEGMENT_SIZE) {        // calc size for writing info_segment
                sz = INFO_SEGMENT_SIZE;
            } else {
                sz = size-INFO_SEGMENT_SIZE*2;
            }
            if (WrFlashSegment((char*)n, (ulong)SEGMENT_B, sz)) {       // write next segment
                return 4;                                               // error, return code
            }
            if (size > INFO_SEGMENT_SIZE*3) {                           // data exists to write
                n = n + INFO_SEGMENT_SIZE;                              // update pointer for next segment write
                sz = size-INFO_SEGMENT_SIZE*3;                          // calc size for writing info_segment
                if (WrFlashSegment((char*)n, (ulong)SEGMENT_A, sz)) {   // write last segment
                    return 5;                                           // error, return code
                }
            }
        }
    }
    return 0;                                               // success
}


// write data to any segment in flash
int WrFlash(char* data, ulong segment, uint size) __monitor {
    if (size > SEGMENT_SIZE) {                              // error, size too big
        return 1;                                           // error, return code
    }
    if (WrFlashSegment(data, segment, size)) {              // validate write
        return 2;                                           // error, return  code
    }
    return 0;                                               // success
}
