#ifndef global_h
#define global_h

#include <__cross_studio_io.h>
#include <msp430x26x.h>
#include <inttypes.h>
#include "bit.h"
#include "types.h"


// macros
#define DisableWDT (IE1 &= ~WDTIE)                  // disable watchdog interrupt
#define EnableWDT (IE1 |= WDTIE)                    // enable watchdog interrupt

// clocks & frequencies
#define MCLK_FREQ       10.0e6
#define SMCLK_FREQ      10.0e6
#define ACLK_FREQ       32768
#define BAUD_9600       9600
#define TA_CLK          5.0e6

//xADC
#define xADC_INT_TIME   8.5e-3                      // ADC integration time (sec) ** MUST be < 13 mSec **
#define LED_ON_DELAY    7                           // number of readings to wait when first switching LED on
#define N_MEASURE       20                          // number of measurements per LED on-time
                                                    // stop measuring ~ 16mSec before color switch

#define STARTUP_DELAY   12                          // number of readings to delay on startup for ratio calcs
#define RATIO_ADJ_INT   5                           // 5-sec intervals for calculating ratio adjust

// LEDs
#define iDARK           0                           // array index
#define iRED            1
#define iGREEN          2
#define iBLUE           3
#define DEF_RED_INT     100.0                       // default LED intensities (fixed)
#define DEF_GRN_INT     100.0
#define DEF_BLUE_INT    100.0

// filter/color count
#define N_COLOR         4                           // number of colors (including dark)
#define N_FILTER        10                          // number of filters (including reference)
#define N_RDG           25                          // maximum number of readings per sample

// unpolled data output
#define MAX_OUT_INT     120                         // maximum output interval (seconds)
#define DEF_OUTPUT_INT  1                           // default output interval of 1 sec
#define MENU_PAUSE      10                          // pause unpolled data after menu display

// low-pass filter
// default low-pass filter constants (4-stage, single pole)
#define N_LPF           5                           // number of low-pass filter constants
#define DEF_LPF0        0.217871821                 // a0
#define DEF_LPF1        1.267184848                 // b1
#define DEF_LPF2        -0.602159039                // b2
#define DEF_LPF3        0.127174468                 // b3
#define DEF_LPF4        -0.010072097                // b4
#define LLVL_LPF        0.8                         // low-level low-pass filtering

// special characters
#define SOH             0x01                        // start of header
#define ETX             0x03                        // end of text
#define BS              0x08                        // backspace
#define CR              0x0D                        // carraige return
#define LF              0x0A                        // line feed
#define DLE             0x10                        // data link escape
#define XON             0x11                        // (DC1 - device control 1) software handshaking chars
#define XOFF            0x13                        // (DC3 - device control 3)

// flags
// lower word of flags for on-board use
#define F_RED           0x00000001L                 // red LED on
#define F_GREEN         0x00000002L                 // green LED on
#define F_BLUE          0x00000004L                 // blue LED on
#define F_DATA_ON       0x00000008L                 // unpolled data on
#define F_DATA_UPDATE   0x00000010L                 // time to generate 1-sec data
#define F_DATA_OUT      0x00000020L                 // time to send data
#define F_CASEHTR_ON    0x00000040L                 // case heater operational
#define F_AUXHTR_ON     0x00000080L                 // aux heater operational
#define F_FPSW_PRESS    0x00000100L                 // front-pannel button pressed
// upper word of flags for output stream use
#define F_FILTER_CHG    0x00010000L                 // filter change
#define F_FLOW_ERR      0x00020000L                 // flow sensor error
#define F_INT_ERR       0x01000000L                 // intensity (LED or filter) error
#define F_TEMP_ERR      0x02000000L                 // temperature error, case or air
#define F_TSTBL_ERR     0x04000000L                 // case temperature not within limits


// heater control
// TB1 (P4.1) = case heater control
// TB2 (P4.2) = auxillary heater
#define TCASE_MIN      -99.99                       // min value for case temp & heater setpt
#define TCASE_MAX       99.99                       // max value for case temp & heater setpt
#define TAIR_MIN       -99.99                       // min value for air temp
#define TAIR_MAX        99.99                       // max value for air temp
#define DEF_HSP         35.0                        // default heater setpt
#define DEF_TCASE_STBL  1.0                         // default stabilized case temperature window
#define TSTBL_MIN       0.05                        // min & max values for case temp stable window
#define TSTBL_MAX       50.0

// default thermistor cal (YSI44031)
#define DEF_T_CAL0      1.039004E-3
#define DEF_T_CAL1      2.376432E-4
#define DEF_T_CAL2      0.0
#define DEF_T_CAL3      1.616100E-7

// default flow sensor cal (AWM43600)
#define DEF_FLOW_CAL0   -0.8037
#define DEF_FLOW_CAL1   1.204
#define DEF_FLOW_CAL2   -0.4994
#define DEF_FLOW_CAL3   0.0974

// min & max flows
#define FLOW_MIN        0.0
#define FLOW_MAX        3.0
#define FLOW_SCALING    0.1                         // scaling for local display

// serial number size, & default serial number
#define N_SN            10
#define DEF_SN          "10.000\0"

// P5.0 = front-panel push button with pull-up (1=not pressed, 0=pressed)
#define FPSW_PIN        P5IN
#define FPSW            0x01
#define FPSW_EVENT      2                           // take at least 2 seconds for front-panel push-button events

// detector intensity checks
#define TOO_BRIGHT      1.0e+6
#define TOO_DARK        10.0e+3
#define MAX_ERR_CNT     10


enum LED_MODE {AUTO, FIXED};                        // brightness setting modes

typedef struct {
    ulong       cnt[N_COLOR];
    float       val[N_COLOR][N_FILTER];
} AvgStruct;

typedef struct {
    ulong       cnt;
    float       val[N_COLOR][N_FILTER];
    float       Tc, Ta, Flw;
} OutAvgStruct;



#endif
