#include <stdio.h>
#include <string.h>
#include <math.h>
#include "global.h"
#include "clock.h"
#include "sio.h"
#include "adc12.h"
#include "dac12.h"
#include "config.h"
#include "xadc.h"
#include "led.h"
#include "heater.h"
#include "valve.h"

const char      VERSION[7] = "10.105\0";
char            szStr[200];
CfgStruct       cfg;                                // config/cal data
int             i,j,k;
volatile uint   sampleCnt;                          // number of xADC samples on current LED
volatile ulong  flags;                              // gloabal flags
volatile ulong  elapsedTime;                        // elapsed time from startup (seconds)
uint            pauseCnt;                           // menu pause counter
ulong           PBevent,                            // elapsed time when filter change button pressed
                ul[N_COLOR];                        // used to convert float data to hex ascii
long int        bin[N_FILTER];                      // binary xADC data for each filter
AvgStruct       detAvg;                             // averaging data for 1-sec values
OutAvgStruct    outAvg;                             // averaging data for output
float           f[N_LPF];
float           detVal[N_COLOR][N_FILTER][N_LPF];   // detector low-pass filter data
float           Tc, Ta, Flw;                        // current data from ADC12
float           iFlow;                              // integrated flow (m3)
uint            Ierr[N_COLOR-1][N_FILTER];          // intensity error counter

// data located at the top of the data area to catch a stack overflow
// assign location = ___end_UDATA0
// after build, check map file, then set and rebuild
#define STACK_SENT_VAL 0xdeadbeef
ulong StackSentinal __at 0x1A36;

// function prototypes
void WatchdogInit (void);                           // init watchdog timer used for stack monitoring
void PortsInit (void);                              // init I/O ports
void SvsInit (void);                                // init Supply Voltage Supervisor for AVcc


void main(void)
{
    DisableWDT;                                     // start with WDT disabled
    PortsInit();                                    // setup ports & pull up/down
    ClockInit();                                    // setup clock source
    WatchdogInit();                                 // WDT interrupt disabled by default
    EnableWDT;                                      // enable watchdog
    elapsedTime = 0;                                // init times
    pauseCnt = 0;
    PBevent = 0;
    sampleCnt = 0;                                  // init LED sample count
    Tc = 0.0;                                       // init ADC12 averaging data
    Ta = 0.0;
    Flw = 0.0;
    iFlow = 0.0;                                    // reset total flow
    SvsInit();                                      // init Supply Voltage Supervisor for AVcc
    TimerA_Init();                                  // init timers
    TimerB_Init();
    SIOinit(57600);                                 // init serial port
    _EINT();                                        // enable global interrupts
    ADC12Init();                                    // Init & start ADC
    DAC12Init();
    SIOtxMsg("");
    SIOtxStr("NOAA/ESRL/GMD C3W - ");
    SIOtxMsg((char*)VERSION);
    if (RdConfig(&cfg) != 0) {                      // error reading config/cals
        DefaultCfg(&cfg);                           // init config/cal to default values
        SIOtxMsg("Read config error, using defaults");
        WrConfig(&cfg);                             // save new config/cal info to flash
    }
    InitHeater(CASE_HTR, cfg.p, cfg.i, cfg.d);      // init heater PID data
    InitHeater(AUX_HTR, cfg.p, cfg.i, cfg.d);
    flags |= F_DATA_ON;                             // start with unpolled data on
    OpenValve(0);                                   // start with all valves off
    cfg.filter_ch = 0;
    WrConfig(&cfg);                                 // save new config/cal info to flash

    for (i=0;i<N_COLOR;i++) {                       // init data
        for (j=0;j<N_FILTER;j++) {
            for (k=0;k<N_LPF;k++) {
                detVal[i][j][k] = 0.0;              // data history for LPF
            }
            detAvg.val[i][j] = 0.0;                 // data for 1-sec averaging
            outAvg.val[i][j] = 0.0;                 // data for output averaging
        }
        detAvg.cnt[i] = 0;
    }
    outAvg.cnt = 0;
    outAvg.Tc = 0.0;
    outAvg.Ta = 0.0;
    outAvg.Flw = 0.0;
    for (i=0;i<N_COLOR-1;i++) {                     // init error count
        for (j=0;j<N_FILTER;j++) {
            Ierr[i][j] = 0;
        }
    }

    while(1) {                                                              // run forever
        i = SIOprocMsg(szStr);
        if (flags & F_DATA_UPDATE) {                                        // time to update 1-sec data
            flags &= ~F_DATA_UPDATE;                                        // reset flag
            Tc = Tcase();                                                   // update data from ADC12
            Ta = Tair();
            Flw = Flow();
            DAC12set(Flw*FLOW_SCALING);                                     // update display
            outAvg.cnt++;                                                   // update averaging data
            outAvg.Tc += Tc;
            outAvg.Ta += Ta;
            outAvg.Flw += Flw;
            if ((Tc > TCASE_MIN) && (Tc < TCASE_MAX)) {                     // validate case temperature
                SetHeater(CASE_HTR, UpdatePID(CASE_HTR, cfg.Tcase.Setpt, Tc));      // update heater control
            }
            // average data, 1-sec
            // prepare LP filter data for newset point
            for (i=0;i<N_COLOR;i++) {                                       // push LP filter buffer
                for (j=0;j<N_FILTER;j++) {
                    for (k=N_LPF-1;k>0;k--) {
                        detVal[i][j][k] = detVal[i][j][k-1];
                    }
                }
            }

            // calc 1-sec average
            _DINT();                                                        // disable interrupts while calculating 1-sec averages
            for (i=0;i<N_COLOR;i++) {
                if (detAvg.cnt[i] > 0) {                                    // prevent div by 0
                    for (j=0;j<N_FILTER;j++) {
                        detVal[i][j][0] = detAvg.val[i][j] / (float)detAvg.cnt[i];      // calc 1-sec average
                        detAvg.val[i][j] = 0.0;                                         // reset summation
                    }
                    detAvg.cnt[i] = 0;                                      // reset counter
                } else {                                                    // no data to average
                    for (j=0;j<N_FILTER;j++) {
                        detVal[i][j][0] = 0.0;
                    }
                }
            }
            _EINT();                                                        // finished 1-sec averaging, enable interrupts

            // apply filter
            for (i=0;i<N_COLOR;i++) {                                       // push LP filter buffer
                for (j=0;j<N_FILTER;j++) {
                    f[0] = detVal[i][j][0] * cfg.lpf[0];
                    f[1] = detVal[i][j][1] * cfg.lpf[1];
                    f[2] = detVal[i][j][2] * cfg.lpf[2];
                    f[3] = detVal[i][j][3] * cfg.lpf[3];
                    f[4] = detVal[i][j][4] * cfg.lpf[4];
                    detVal[i][j][0] = f[0] + f[1] + f[2] + f[3] + f[4];
                }
            }

            // update summation data for output averaging
            for (i=0;i<N_COLOR;i++) {
                for (j=0;j<N_FILTER;j++) {
                    outAvg.val[i][j] += detVal[i][j][0];
                }
            }

            // calculate integrated flow
            if (!(flags & F_FILTER_CHG) && (cfg.filter_ch != 0)) {          // filter in use
                iFlow += Flw * 16.66667e-6;                                 // total flow through spot (1.0 l = 0.001 m3)
            }

            // average data
            if (flags & F_DATA_OUT) {                                       // time to output average data
                flags &= ~F_DATA_OUT;                                       // reset flag
                if (outAvg.cnt > 0) {                                       // prevent div by 0
                    for (i=0;i<N_COLOR;i++) {                               // average data
                        for (j=0;j<N_FILTER;j++) {
                            outAvg.val[i][j] /= (float)outAvg.cnt;
                        }
                    }
                    f[0] = (float)outAvg.cnt;                               // format conversion to temp var
                    outAvg.Tc = outAvg.Tc / f[0];                           // calc averages
                    outAvg.Ta = outAvg.Ta / f[0];
                    outAvg.Flw = outAvg.Flw / f[0];
                    if (fabs(outAvg.Tc - cfg.Tcase.Setpt) > cfg.Tstbl) {    // check case temperature stability
                        flags |= F_TSTBL_ERR;                               // temp not stable, set flag
                    } else {
                        flags &= ~F_TSTBL_ERR;                              // temp stable, clear flag
                    }
                } else {                                                    // no data to average
                    for (i=0;i<N_COLOR;i++) {
                        for (j=0;j<N_FILTER;j++) {
                            outAvg.val[i][j] = 0.0;
                        }
                    }
                    outAvg.Tc = 0.0;
                    outAvg.Ta = 0.0;
                    outAvg.Flw = 0.0;
                    outAvg.cnt = 0;
                }

                if (pauseCnt) pauseCnt--;                                   // update menu pause counter
                if ((flags & F_DATA_ON)&&(pauseCnt == 0)) {                 // send data

/* decimal output no longer used 20100414
                    snprintf(szStr,sizeof(szStr),"01, %04x, %08lx, %04x, %02x,%6.3f,%9.5,%6.2f,%6.2f",
                            (uint)(flags >> 16),
                            elapsedTime,
                            cfg.filter_id,
                            cfg.filter_ch,
                            outAvg.Flw,
                            iFlow,
                            outAvg.Tc,
                            outAvg.Ta);
                    SIOtxStr(szStr);
                    for (i=0;i<N_FILTER;i++) {
                        snprintf(szStr,sizeof(szStr),",%9.0f,%9.0f,%9.0f,%9.0f",
                            outAvg.val[iDARK][i] * 100.0,
                            outAvg.val[iRED][i] * 100.0,
                            outAvg.val[iGREEN][i] * 100.0,
                            outAvg.val[iBLUE][i] * 100.0);
                        SIOtxStr(szStr);
                    }
                    SIOtxStr("\r\n");
*/
                    SIOtxStr("03,");
                    snprintf(szStr,sizeof(szStr)-1," %04x, %08lx,",(uint)(flags >> 16),elapsedTime);
                    SIOtxStr(szStr);
                    snprintf(szStr,sizeof(szStr)-1," %04x, %02x,",cfg.filter_id,cfg.filter_ch);
                    SIOtxStr(szStr);
                    snprintf(szStr,sizeof(szStr)-1,"%6.3f,%9.5f,%6.2f,%6.2f",outAvg.Flw,iFlow,outAvg.Tc,outAvg.Ta);
                    SIOtxStr(szStr);
                    for (i=0;i<N_FILTER;i++) {
                        memcpy(&ul[iDARK],&outAvg.val[iDARK][i],sizeof(outAvg.val[iDARK][i]));
                        memcpy(&ul[iRED],&outAvg.val[iRED][i],sizeof(outAvg.val[iRED][i]));
                        memcpy(&ul[iGREEN],&outAvg.val[iGREEN][i],sizeof(outAvg.val[iGREEN][i]));
                        memcpy(&ul[iBLUE],&outAvg.val[iBLUE][i],sizeof(outAvg.val[iBLUE][i]));
                        snprintf(szStr,sizeof(szStr)-1,", %08lx, %08lx, %08lx, %08lx",ul[iDARK],ul[iRED],ul[iGREEN],ul[iBLUE]);
                        SIOtxStr(szStr);
                    }
                    SIOtxStr("\r\n");
                }
                for (i=0;i<N_COLOR;i++) {                               // reset output averaging data
                    for (j=0;j<N_FILTER;j++) {
                        outAvg.val[i][j] = 0.0;
                    }
                }
                outAvg.Tc = 0.0;                                        // reset for averaging
                outAvg.Ta = 0.0;
                outAvg.Flw = 0.0;
                outAvg.cnt = 0;
            }

            // check for intensity error
            if (!(flags & F_FILTER_CHG)) {                              // check for intensity errors
                for (i=0;i<N_COLOR-1;i++) {                             // check RGB for out-of-range signal
                    for (j=0;j<N_FILTER;j++) {                          // check each detector
                        if (detVal[i+1][j][0] > TOO_BRIGHT) {           // det signal too high
                            Ierr[i][j]++;                               // update error counter
                        } else {
                            if (detVal[i+1][j][0] < TOO_DARK) {         // det signal too low
                                Ierr[i][j]++;                           // update error counter
                            } else {                                    // no error
                                Ierr[i][j] = 0;                         // reset counter
                            }
                        }
                        if (Ierr[i][j] > MAX_ERR_CNT) {                 // error count sufficient
                            flags |= F_INT_ERR;                         // set intensity error flag
                        }
                    }
                }
            }

            // check intensity error status
            if ((flags & F_INT_ERR) && !(flags & F_FILTER_CHG)) {       // error, flash front-panel LED
                FPLED_POUT ^= FPLED;                                    // toggle LED
            }

            // process button press
            if (flags & F_FPSW_PRESS) {
                if (elapsedTime - PBevent >= FPSW_EVENT) {              // button pressed allowed
                    PBevent = elapsedTime;                              // mark time of button press
                    OpenValve(0);                                       // all valves off
                    cfg.filter_ch = 0;                                  // no spot active
                    if (!(flags & F_FILTER_CHG)) {                      // only if not currently changing filter
                        flags |= F_FILTER_CHG;                          // set filter change flag
                        flags &= ~F_INT_ERR;                            // clear intensity error flag
                        SetFpLED(1);                                    // turn on LED
                        for (i=0;i<N_COLOR-1;i++) {                     // reset intensity error count
                            for (j=0;j<N_FILTER;j++) {
                                Ierr[i][j] = 0;
                            }
                        }
                    } else {                                            // currently changing filter
                        flags &= ~F_FILTER_CHG;                         // clear filter change flag
                        SetFpLED(0);                                    // turn off LED
                        cfg.filter_id++;                                // new filter ID
                        iFlow = 0.0;                                    // reset integrated flow
                    }
                    WrConfig(&cfg);                                     // save new config/cal info to flash
                }
                flags &= ~F_FPSW_PRESS;                                 // clear flag (processed)
            }

        }
    }
}


// init Supply Voltage Supervisor for AVcc
void SvsInit (void) {
    SVSCTL = VLD3+VLD1;                             // monitor AVcc, AVcc < 3.05 sets flag, does not cause POR
}


// Initialize watchdog, used to detect stack overflow
void WatchdogInit(void) {
    DisableWDT;                                                 // disable WDT
    WDTCTL = WDTPW + WDTTMSEL + WDTCNTCL + WDTSSEL;             // WDT is interval timer using ACLK/32768 = 32.768KHz/32768 = 1/sec
    StackSentinal = STACK_SENT_VAL;                             // initialize stack sentinal
}


// init I/O ports
void PortsInit(void) {
    P1OUT = 0x04;                   // -Dxmit high at startup
    P1DIR = 0xDD;                   // (11011101) set direction: 0=in, 1=out
    P1REN = 0x20;                   // (00100000) pull-up/down resistor: 0=disabled, 1=enabled
    P1SEL = 0x58;                   // (01011000) peripheral assigned to pin
    P1IES = 0x02;                   // interrupt edge select, 0= _- , 1= -_
    P1IE = 0x02;                    // enable interrupt for -DValid signal
    P2OUT = 0;
    P2DIR = 0xE0;                   // (11100000) set direction: 0=in, 1=out
    P2REN = 0;                      // (00000000) pull-up/down resistor: 0=disabled, 1=enabled
    P2SEL = 0;                      // (00000000) peripheral assigned to pin
    P3OUT = 0;
    P3DIR = 0xDF;                   // (11011111) set direction: 0=in, 1=out
    P3REN = 0;                      // (00000000) pull-up/down resistor: 0=disabled, 1=enabled
    P3SEL = 0x30;                   // (00110000) peripheral assigned to pin
    P4OUT = 0;
    P4DIR = 0x3E;                   // (00111110) set direction: 0=in, 1=out
    P4REN = 0xC1;                   // (11000001) pull-up/down resistor: 0=disabled, 1=enabled
    P4SEL = 0x06;                   // (00000110) peripheral assigned to pin
    P5OUT = 0xFD;
    P5DIR = 0x02;                   // (00000010) set direction: 0=in, 1=out
    P5REN = 0xFD;                   // (11111101) pull-up/down resistor: 0=disabled, 1=enabled
    P5SEL = 0;                      // (00000000) peripheral assigned to pin
    P6OUT = 0;
    P6DIR = 0xF8;                   // (11111000) set direction: 0=in, 1=out
    P6REN = 0;                      // (00000000) pull-up/down resistor: 0=disabled, 1=enabled
    P6SEL = 0x07;                   // (00000111) peripheral assigned to pin
}


// Check for stack overflow
void watchdog_timer (void) __interrupt[WDT_VECTOR] {

    if (StackSentinal != STACK_SENT_VAL) {                  // stack overflow
        FCTL1 = 0;                                          // force reset
    }
}


// PORT1 interrupt service routine
void Port1_ISR (void) __interrupt[PORT1_VECTOR] {
    int         n,iColor;
    ulong       led;

    if (P1IFG & xADC_DVALID) {                              // xADC conversion complete
        P1IFG = 0;                                          // reset all PORT1 interrupt flags
        _EINT();                                            // enable interrupts
        Rd_xADC(bin);                                       // get xADC data for all filters
        led = flags & (F_RED + F_GREEN + F_BLUE);           // save LED state when data was collected

        // process xADC data
        switch (led) {                                      // compute color index for array access
            case F_RED:                                     // red LED is on
                iColor = iRED;
                break;
            case F_GREEN:                                   // green LED is on
                iColor = iGREEN;
                break;
            case F_BLUE:                                    // blue LED is on
                iColor = iBLUE;
                break;
            default:                                        // no LEDs on, must be dark
                iColor = iDARK;
        }
        // don't used data until LED output is stable and stop before LED turns off
        if ((sampleCnt >= LED_ON_DELAY) && (sampleCnt < LED_ON_DELAY+N_MEASURE)) {
            for (n=0;n<N_FILTER;n++) {                      // push LP filter buffer
                detAvg.val[iColor][n] += detVal[iColor][n][0]*(1.0 - LLVL_LPF) + (float)bin[n]*LLVL_LPF;       // update summation
            }
            detAvg.cnt[iColor]++;                           // update summation counter for averaging
        }
        sampleCnt++;                                        // update sample count number
    } else {                                                // unspecified pin caused interrupt (shouldn't happen)
        P1IFG = 0;                                          // reset all PORT1 interrupt flags
    }
}
