#include "sio.h"
#include "global.h"
#include "config.h"
#include "clock.h"
#include "heater.h"
#include "valve.h"
#include "led.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

extern char             VERSION[];                      // main.c
extern volatile ulong   flags;                          // main.c - global flags
extern CfgStruct        cfg;                            // main.c - global configuration/cal data
extern AvgStruct        outAvg;                         // main.c - output averaging data
extern uint             outCnt;                         // clock.c - number of 1-sec counts for output interval
extern ulong            elapsedTime;                    // main.c - elapsed time (sec)
extern uint             pauseCnt;                       // main.c - pause counter
extern ulong            PBevent;                        // main.c - button press event time
extern float            iFlow;                          // main.c - integrate flow (m3)
extern uint             Ierr[N_COLOR-1][N_FILTER];      // main.c - intensity error count

typedef struct {                        // receive queue structure
    char    buffer[MSG_RX_QUEUE_SIZE];
    uint    first,last;
    uchar   msg_avail;
} SIOrxQueueStruct;

typedef struct {                        // transmit queue structure
    char    buffer[MSG_TX_QUEUE_SIZE];
    uint    first, last;
} SIOtxQueueStruct;

enum ALLMENUS_ID {MENU, MENU_SC, MAIN, HIDE, SHOW};
#define ALLMENU_CNT     5

enum MAINMENU_ID {CFG, CAL, FLT_STOP, FLT_START, FLT_SPOT};
#define MAINMENU_CNT    5

enum CFGMENU_ID {OUTINT, SN, SETSN, FW, CFG_LPF, HSP, STBL, RESET};
#define CFGMENU_CNT     8

enum CALMENU_ID {CASE, AIR, FLOW, PID, LED};
#define CALMENU_CNT     5

const char AllMenus_cmd[ALLMENU_CNT][6] = {"menu",                  // show main
                                           "?",                     // alternate show menu
                                           "main",                  // switch to main menu
                                           "hide",                  // stop unpolled data
                                           "show"};                 // enable unpolled data

const char MainMenu_cmd[MAINMENU_CNT][6] = {"cfg",                  // configuration settings
                                            "cal",                  // calibration data
                                            "stop",                 // stop filter, start filter change
                                            "go",                   // start new filter
                                            "spot"};                // activate specified filter spot (0=none)

const char CfgMenu_cmd[CFGMENU_CNT][6] =  {"oint",                  // show/set output interval (seconds)
                                           "sn",                    // serial number (read-only)
                                           "setsn",                 // hidden command to set the serial number
                                           "fw",                    // firmware version (read-only)
                                           "lpf",                   // low-pass filter
                                           "hsp",                   // case heater setpt (C)
                                           "stbl",                  // case temperature stable window
                                           "rst"};                  // reset cfg and cals

const char CalMenu_cmd[CALMENU_CNT][6] =  {"tc",                    // show/set case temperature cals
                                           "ta",                    // show/set air temperature cals
                                           "flow",                  // show/set flow cals
                                           "pid",                   // PID gain settings
                                           "led"};                  // LED mode and intensity settings

const char  delims[8] = {" ,=[]\n\r\0"};

enum MENU {MAIN_MENU, CFG_MENU, CAL_MENU};
uint    currentMenu = MAIN_MENU;                // keep track of menu level

volatile SIOrxQueueStruct    SIOrxQ;            // receive and transmit queues
volatile SIOtxQueueStruct    SIOtxQ;


void SIOinitSerial (uint baudrate);             // initialize the serial port
uchar SIOdisableTxInt (void);                   // manage transmit interrupts (UART1)
void SIOenableTxInt (void);
void SIOsetTxInt (uchar state);
uchar SIOdisableRxInt (void);                   // manage receive interrupts (UART1)
void SIOenableRxInt (void);
void SIOsetRxInt (uchar state);
uint SIOtxQspace (void);                        // return the space available in the tranmit queue
void SIOrxInt (void);                           // receive interrupt handler
void SIOtxInt (void);                           // transmit interrupt handler
void DispMenu (void);                           // display current menu


// initialize the UART serial port (N,8,1)
void SIOinitSerial (uint baudrate) {
    uint    i;
    float   n;

    SIOdisableTxInt();                  // disable interrupts
    SIOdisableRxInt();
    UCA0CTL1 = UCSWRST;                 // disable UART
    UCA0CTL1 |= UCSSEL1;                // baudrate generator uses SMCLK
    UCA0CTL0 |= 0;                      // 8 data, no parity, 1 stop bit, UART mode
    UCA0MCTL = UCOS16;                  // oversampling enabled
    n = (float)SMCLK_FREQ / (float)baudrate;            // baudrate clk divisor and modualtor
    i = (uint)n / 16;
    UCA0BR0 = i & 0xff;
    UCA0BR1 = i >> 8;
    i = (uint)((((n/16.0) - (float)i)*16.0)+0.5);       // baudrate clk modulator
    UCA0MCTL |= (i & 0x0f) << 4;
    P3DIR |= 0x10;                      // UART Tx = ouput
    P3DIR &= ~0x20;                     // UART Rx = input
    P3SEL |= (0x30);                    // enable UART pins: Tx (P3.4), Rx (P3.5)
    UCA0CTL1 &= ~UCSWRST;               // enable UART
    SIOenableRxInt();                   // enable receive interrupts only (nothing to send yet)
}


// disable Tx interrupt, return previous state
uchar SIOdisableTxInt(void) {
    uchar state;

    state = IE2;                        // save state of transmit interrupt
    IE2 &= ~UCA0TXIE;                   // disable transmit interrupt
    return state;
}


// enable Tx interrupt
void SIOenableTxInt(void) {
    IE2 |= UCA0TXIE;                    // enable transmit interrupt
}


// restore Tx interrupt to specified state
void SIOsetTxInt(uchar state) {
    IE2 |= (state & UCA0TXIE);          // restore state of transmit interrupt
}


// disable Rx interrupt, return previous state
uchar SIOdisableRxInt(void) {
    uchar state;

    state = IE2;                        // save state of receive interrupt
    IE2 &= ~UCA0RXIE;                   // disable receive interrupt
    return state;
}


// enable Rx interrupt
void SIOenableRxInt(void) {
    IE2 |= UCA0RXIE;                    // enable receive interrupt
}


// restore Rx interrupt to specified state
void SIOsetRxInt(uchar state) {
    IE2 |= (state & UCA0RXIE);          // restore state of receive interrupt
}


// return the space available in the tranmit queue
uint SIOtxQspace (void) __monitor {
    uint    used;

    if (SIOtxQ.last < SIOtxQ.first) {                                   // buffer wrap
        used = SIOtxQ.last + MSG_TX_QUEUE_SIZE - SIOtxQ.first;
    } else {                                                            // not wrapped
        used = SIOtxQ.last - SIOtxQ.first;
    }
    return MSG_TX_QUEUE_SIZE - used - 1;
}


// add char to transmit queue
void SIOtxChar (char c) __monitor {
    SIOtxQ.buffer[SIOtxQ.last] = c;                                     // put char in queue
    SIOtxQ.last = ++SIOtxQ.last % MSG_TX_QUEUE_SIZE;                    // update pointer
    if (IFG2 & UCA0TXIFG) {                                             // Tx buffer empty
        UCA0TXBUF = SIOtxQ.buffer[SIOtxQ.first];                        // transmit the first char
        SIOtxQ.first = ++SIOtxQ.first % MSG_TX_QUEUE_SIZE;              // update index pointer
        SIOenableTxInt();                                               // enable transmit irq
    }
}


// initalize SIO, setup serial port and receive and transmit queues
void SIOinit (uint baudrate) {
    SIOrxQ.first = 0;                       // init queues
    SIOrxQ.last = 0;
    SIOrxQ.msg_avail = 0;
    SIOtxQ.first = 0;
    SIOtxQ.last = 0;
    SIOinitSerial(baudrate);                // setup serial port UART1
}


// send a string
void SIOtxStr (char* str) {
    uint   i,len;

    len = strlen(str);
    while (SIOtxQspace() < len) {           // wait for room in queue
    }
    for (i=0;i<len;i++) {                   // add each char
        SIOtxChar(str[i]);
    }
}

// send a string, automatically add CRLF
void SIOtxMsg (char* msg) {
    uint   i,len;

    len = strlen(msg)+2;
    while (SIOtxQspace() < len) {           // wait for room in queue
    }
    for (i=0;i<len-2;i++) {                   // add each char
        SIOtxChar(msg[i]);
    }
    SIOtxChar('\r');
    SIOtxChar('\n');
}


// return true if a message is available in the queue
bool SIOmsgAvail (void) {
    return (SIOrxQ.msg_avail > 0);
}


// process a single message if available in the receive queue
// return is non-zero if a message was processed and returning info in "msg"
uchar SIOprocMsg (char msg[]) {
    int     i,j,k;
    float   f[5];
    char    str[100];
    char*   cmd;
    uchar   uc;
    bool    processed;

    msg[0] = '\0';                              // init return message
    if (SIOrxQ.msg_avail > 0) {                 // validate message availability
        if (SIOrxQ.first != SIOrxQ.last) {      // validate data in queue
            // copy message from queue
            i = 0;                              // init index counter
            while ((SIOrxQ.buffer[(SIOrxQ.first+i)%MSG_RX_QUEUE_SIZE] != CR) && ((SIOrxQ.first+i)%MSG_RX_QUEUE_SIZE != SIOrxQ.last)) {
                msg[i] = tolower(SIOrxQ.buffer[(SIOrxQ.first + i) % MSG_RX_QUEUE_SIZE]);      // convert to lower case
                i++;
            }
            msg[i] = '\0';                                              // terminate string
            SIOrxQ.first = (SIOrxQ.first + i) % MSG_RX_QUEUE_SIZE;      // message removed from queue
            SIOrxQ.msg_avail--;                                         // update msg counter
            while ((SIOrxQ.buffer[SIOrxQ.first] == CR) && (SIOrxQ.first != SIOrxQ.last)) {                            // remove CR-LF
                SIOrxQ.first = (SIOrxQ.first+1) % MSG_RX_QUEUE_SIZE;    // move past terminating CR in queue
            }
            if (SIOrxQ.first == SIOrxQ.last) {                          // nothing left in buffer
                SIOrxQ.msg_avail = 0;                                   // make sure msg available counter is accurate
            }
            processed = false;
            cmd = strtok(msg,delims);                                   // parse command
            // check commands available to all menues first
            for (i=0;i<ALLMENU_CNT;i++) {                               // check for each possible command
                if (strcmp(cmd,AllMenus_cmd[i]) == 0) {                 // command found
                    switch (i) {
                        case MENU:                                      // display main menu
                        case MENU_SC:
                            DispMenu();
                            processed = true;
                            msg[0] = '\0';                              // clear return message
                            break;
                        case MAIN:                                      // goto main menu
                            currentMenu = MAIN_MENU;                    // update menu status
                            DispMenu();                                 // display current menu
                            processed = true;
                            msg[0] = '\0';                              // clear return message
                            break;
                        case HIDE:                                      // stop unpolled data
                            flags &= ~F_DATA_ON;                        // stop unpolled data
                            SIOtxMsg("Unpolled data stopped");
                            processed = true;
                            msg[0] = '\0';                              // clear return message
                            break;
                        case SHOW:                                      // start unpolled data
                            pauseCnt = 0;                               // cancel menu pause
                            flags |= F_DATA_ON;                         // unpolled data started
                            SIOtxMsg("Unpolled data enabled");
                            _DINT();                                    // block interrupts while changing var
                            outCnt = 0;                                 // reset output counter
                            _EINT();                                    // enable interrupts
                            for (j=0;j<N_COLOR;j++) {                   // reset summation data for ouput averaging
                                for (k=0;k<N_FILTER;k++) {
                                    outAvg.val[j][k] = 0.0;
                                }
                                outAvg.cnt[j] = 0;                      // reset output averaging counter
                            }
                            processed = true;
                            msg[0] = '\0';                              // clear return message
                            break;
                    }
                }
            }
            if (!processed) {
                // check for commands relative to current menu
                switch (currentMenu) {
                    case MAIN_MENU:
                        for (i=0;i<MAINMENU_CNT;i++) {                      // check for each possible command
                            if (strcmp(cmd,MainMenu_cmd[i]) == 0) {         // command found
                                switch (i) {
                                    case CFG:                               // configuration info
                                        currentMenu = CFG_MENU;             // update menu status
                                        DispMenu();                         // display current menu
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                    case CAL:                               // calibration data
                                        currentMenu = CAL_MENU;             // update menu status
                                        DispMenu();                         // display current menu
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                    case FLT_STOP:                          // stop filter
                                        OpenValve(0);                       // all valves off
                                        SetFpLED(1);                        // turn on front-panel LED
                                        flags |= F_FILTER_CHG;              // set flag
                                        flags &= ~F_INT_ERR;                // clear intensity error flag
                                        for (j=0;j<N_COLOR-1;j++) {         // reset intensity error count
                                            for (k=0;k<N_FILTER;k++) {
                                                Ierr[j][k] = 0;
                                            }
                                        }
                                        cfg.filter_ch = 0;                  // no spot active
                                        WrConfig(&cfg);                     // save new config/cal info to flash
                                        PBevent = elapsedTime;              // mark psuedo-button press event time
                                        SIOtxMsg("filter stop");            // display status
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                    case FLT_START:                         // filter changed
                                        SetFpLED(0);                        // turn off front-panel LED
                                        flags &= ~F_FPSW_PRESS;             // clear flag, in case the filter change was started by the front-panel button
                                        flags &= ~F_FILTER_CHG;             // clear flag
                                        flags |= F_DATA_ON;                 // automatically start unpolled data
                                        cmd = strtok(NULL,delims);          // parse value
                                        if (cmd != NULL) {                  // new filter id specified
                                            j = (uint)atoi(cmd);                    // convert number
                                            if ((j >= 0) && (j <= 0xFFFF)) {        // validate filter number
                                                cfg.filter_id = j;          // new filter ID
                                            } else {
                                                cfg.filter_id++;            // invalid filter ID, increment instead
                                            }
                                        } else {
                                            cfg.filter_id++;                // new filter ID not specified, increment
                                        }
                                        cfg.filter_ch = 0;                  // start with no filter channel/spot
                                        WrConfig(&cfg);                     // save new config/cal info to flash
                                        iFlow = 0.0;                        // reset integrated flow
                                        SIOtxMsg("filter start");           // display status
                                        pauseCnt = 0;                       // cancel menu pause
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                    case FLT_SPOT:                          // activate specified filter spot, 0-8 (0 = none)
                                        cmd = strtok(NULL,delims);          // parse value
                                        if (cmd != NULL) {                  // value specified
                                            if (strcmp(cmd,"+") == 0) {                 // advance to next spot
                                                cfg.filter_ch = ++cfg.filter_ch % (N_FILTER-1);     // next spot
                                                OpenValve(cfg.filter_ch);               // open/close appropriate valves
                                                WrConfig(&cfg);                         // save new config/cal info to flash
                                                iFlow = 0.0;                            // reset integrated flow
                                                flags |= F_DATA_ON;                     // automatically start unpolled data
                                            } else {                                    // use specified spot
                                                j = (uint)atoi(cmd);                    // convert number
                                                if ((j >= 0) && (j < N_FILTER-1) && (j != cfg.filter_ch)) {     // validate spot number
                                                    cfg.filter_ch = j;                  // set new filter spot
                                                    OpenValve(cfg.filter_ch);           // open appropriate valve
                                                    WrConfig(&cfg);                     // save new config/cal info to flash
                                                    iFlow = 0.0;                        // reset integrated flow
                                                    flags |= F_DATA_ON;                 // automatically start unpolled data
                                                }
                                            }
                                        }
                                        memset(str,0,sizeof(str));          // init string
                                        snprintf(str,sizeof(str)-1,"spot =  %d",cfg.filter_ch);
                                        SIOtxMsg(str);                      // display active channel/spot
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                }
                            }
                        }
                        break;
                    case CFG_MENU:
                        for (i=0;i<CFGMENU_CNT;i++) {                       // check for each possible command
                            if (strcmp(cmd,CfgMenu_cmd[i]) == 0) {          // command found
                                switch (i) {
                                    case OUTINT:                            // show/set output interval (seconds)
                                        cmd = strtok(NULL,delims);          // parse value
                                        if (cmd != NULL) {                  // value specified
                                            j = (uint)atoi(cmd);            // convert number
                                            if ((j > 0) && (j <= MAX_OUT_INT)) {        // validate interval
                                                cfg.outputInterval = j;
                                                WrConfig(&cfg);                         // save new config/cal info to flash
                                            }
                                        }
                                        memset(str,0,sizeof(str));          // init string
                                        snprintf(str,sizeof(str)-1,"output interval = %d",cfg.outputInterval);
                                        SIOtxMsg(str);                      // display new cal values
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                    case SN:                                // read serial number
                                        SIOtxStr("serial number = ");
                                        SIOtxStr(cfg.sn);
                                        SIOtxStr("\r\n");
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                    case SETSN:                             // set serial number
                                        cmd = strtok(NULL,delims);          // parse value
                                        if (cmd != NULL) {                  // value specified
                                            strncpy(cfg.sn,cmd,N_SN-1);     // copy serial number
                                            cfg.sn[N_SN-1] = '\0';          // insure null termination
                                            WrConfig(&cfg);                 // save new config/cal info to flash
                                        }
                                        SIOtxStr("serial number = ");
                                        SIOtxStr(cfg.sn);
                                        SIOtxStr("\r\n");
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                    case FW:                                // show firmware version
                                        SIOtxStr("firmware vers = ");
                                        SIOtxStr(VERSION);
                                        SIOtxStr("\r\n");
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                    case CFG_LPF:                           // show/set low-pass filter
                                        cmd = strtok(NULL,delims);                          // parse next value
                                        if (cmd != NULL) {                                  // value specified
                                            f[0] = (float)atof(cmd);                        // convert number
                                            cmd = strtok(NULL,delims);                      // parse next value
                                            if (cmd != NULL) {                              // value specified
                                                f[1] = (float)atof(cmd);                    // convert number
                                                cmd = strtok(NULL,delims);                  // parse next value
                                                if (cmd != NULL) {                          // value specified
                                                    f[2] = (float)atof(cmd);                // convert number
                                                    cmd = strtok(NULL,delims);              // parse next value
                                                    if (cmd != NULL) {                      // value specified
                                                        f[3] = (float)atof(cmd);            // convert number
                                                        cmd = strtok(NULL,delims);          // parse next value
                                                        if (cmd != NULL) {                  // value specified
                                                            f[4] = (float)atof(cmd);        // convert number
                                                            cfg.lpf[4] = f[4];              // update cal value
                                                        }
                                                        cfg.lpf[3] = f[3];                  // update cal value
                                                    }
                                                    cfg.lpf[2] = f[2];                      // update cal value
                                                }
                                                cfg.lpf[1] = f[1];                          // update cal value
                                            }
                                            cfg.lpf[0] = f[0];                              // update cal value
                                            WrConfig(&cfg);                 // save new config/cal info to flash
                                        }
                                        memset(str,0,sizeof(str));          // init string
                                        snprintf(str,sizeof(str)-1,"lpf = %1.6e, %1.6e, %1.6e, %1.6e, %1.6e",cfg.lpf[0],cfg.lpf[1],cfg.lpf[2],cfg.lpf[3],cfg.lpf[4]);
                                        SIOtxMsg(str);                      // display new cal values
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                    case HSP:                               // case heater setpt
                                        cmd = strtok(NULL,delims);                          // parse next value
                                        if (cmd != NULL) {                                  // value specified
                                            f[0] = (float)atof(cmd);                        // convert number
                                            if ((f[0] >= TCASE_MIN) && (f[0] <= TCASE_MAX )) {       // validate new setpt
                                                cfg.Tcase.Setpt = f[0];                     // save new value
                                                WrConfig(&cfg);                             // save new config/cal info to flash
                                            }
                                        }
                                        memset(str,0,sizeof(str));          // init string
                                        snprintf(str,sizeof(str)-1,"hsp = %1.1f",cfg.Tcase.Setpt);
                                        SIOtxMsg(str);                      // display setpt temperature value
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                    case STBL:                              // case temperature stable window (C)
                                        cmd = strtok(NULL,delims);                          // parse next value
                                        if (cmd != NULL) {                                  // value specified
                                            f[0] = (float)atof(cmd);                        // convert number
                                            if ((f[0] >= TSTBL_MIN) && (f[0] <= TSTBL_MAX )) {       // validate new setpt
                                                cfg.Tstbl = f[0];                           // save new value
                                                WrConfig(&cfg);                             // save new config/cal info to flash
                                            }
                                        }
                                        memset(str,0,sizeof(str));          // init string
                                        snprintf(str,sizeof(str)-1,"stbl = %1.1f",cfg.Tstbl);
                                        SIOtxMsg(str);                      // display setpt temperature value
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                    case RESET:                             // reset config/cals to defaults
                                        DefaultCfg(&cfg);
                                        WrConfig(&cfg);                     // save new config/cal info to flash
                                        SIOtxMsg("cfg & cals reset to defaults");
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                }
                            }
                        }
                        break;
                    case CAL_MENU:
                        for (i=0;i<CALMENU_CNT;i++) {                       // check for each possible command
                            if (strcmp(cmd,CalMenu_cmd[i]) == 0) {          // command found
                                switch (i) {
                                    case CASE:                              // show/set case temperature cal
                                        cmd = strtok(NULL,delims);                          // parse next value
                                        if (cmd != NULL) {                                  // value specified
                                            f[0] = (float)atof(cmd);                        // convert number
                                            cmd = strtok(NULL,delims);                      // parse next value
                                            if (cmd != NULL) {                              // value specified
                                                f[1] = (float)atof(cmd);                    // convert number
                                                cmd = strtok(NULL,delims);                  // parse next value
                                                if (cmd != NULL) {                          // value specified
                                                    f[2] = (float)atof(cmd);                // convert number
                                                    cmd = strtok(NULL,delims);              // parse next value
                                                    if (cmd != NULL) {                      // value specified
                                                        f[3] = (float)atof(cmd);            // convert number
                                                        cfg.Tcase.K[3] = f[3];              // update cal value
                                                    }
                                                    cfg.Tcase.K[2] = f[2];                  // update cal value
                                                }
                                                cfg.Tcase.K[1] = f[1];                      // update cal value
                                            }
                                            cfg.Tcase.K[0] = f[0];                          // update cal value
                                            WrConfig(&cfg);                 // save new config/cal info to flash
                                        }
                                        memset(str,0,sizeof(str));          // init string
                                        snprintf(str,sizeof(str)-1,"Tcase = %1.6e, %1.6e, %1.6e, %1.6e",cfg.Tcase.K[0],cfg.Tcase.K[1],cfg.Tcase.K[2],cfg.Tcase.K[3]);
                                        SIOtxMsg(str);                      // display new cal values
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                    case AIR:                               // show/set air temperature cal
                                        cmd = strtok(NULL,delims);                          // parse next value
                                        if (cmd != NULL) {                                  // value specified
                                            f[0] = (float)atof(cmd);                        // convert number
                                            cmd = strtok(NULL,delims);                      // parse next value
                                            if (cmd != NULL) {                              // value specified
                                                f[1] = (float)atof(cmd);                    // convert number
                                                cmd = strtok(NULL,delims);                  // parse next value
                                                if (cmd != NULL) {                          // value specified
                                                    f[2] = (float)atof(cmd);                // convert number
                                                    cmd = strtok(NULL,delims);              // parse next value
                                                    if (cmd != NULL) {                      // value specified
                                                        f[3] = (float)atof(cmd);            // convert number
                                                        cfg.Tair.K[3] = f[3];               // update cal value
                                                    }
                                                    cfg.Tair.K[2] = f[2];                   // update cal value
                                                }
                                                cfg.Tair.K[1] = f[1];                       // update cal value
                                            }
                                            cfg.Tair.K[0] = f[0];                           // update cal value
                                            WrConfig(&cfg);                 // save new config/cal info to flash
                                        }
                                        memset(str,0,sizeof(str));          // init string
                                        snprintf(str,sizeof(str)-1,"Tair = %1.6e, %1.6e, %1.6e, %1.6e",cfg.Tair.K[0],cfg.Tair.K[1],cfg.Tair.K[2],cfg.Tair.K[3]);
                                        SIOtxMsg(str);                      // display new cal values
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                    case FLOW:
                                        cmd = strtok(NULL,delims);          // get settings
                                        if (cmd != NULL) {                  // set cals
                                            f[0] = (float)atof(cmd);                        // convert number
                                            cmd = strtok(NULL,delims);                      // parse next value
                                            if (cmd != NULL) {                              // value specified
                                                f[1] = (float)atof(cmd);                    // convert number
                                                cmd = strtok(NULL,delims);                  // parse next value
                                                if (cmd != NULL) {                          // value specified
                                                    f[2] = (float)atof(cmd);                // convert number
                                                    cmd = strtok(NULL,delims);              // parse next value
                                                    if (cmd != NULL) {                      // value specified
                                                        f[3] = (float)atof(cmd);            // convert number
                                                        cfg.Flow.K[3] = f[3];               // update cal value
                                                    }
                                                    cfg.Flow.K[2] = f[2];                   // update cal value
                                                }
                                                cfg.Flow.K[1] = f[1];                       // update cal value
                                            }
                                            cfg.Flow.K[0] = f[0];                           // update cal value
                                            WrConfig(&cfg);                 // save new config/cal info to flash
                                        }
                                        memset(str,0,sizeof(str));          // init string
                                        snprintf(str,sizeof(str)-1,"Flow = %1.6e, %1.6e, %1.6e, %1.6e",cfg.Flow.K[0],cfg.Flow.K[1],cfg.Flow.K[2],cfg.Flow.K[3]);
                                        SIOtxMsg(str);                      // display new cal values
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                    case PID:                               // PID gain settings
                                        cmd = strtok(NULL,delims);                          // parse next value
                                        if (cmd != NULL) {                                  // value specified
                                            f[0] = (float)atof(cmd);                        // convert number
                                            cmd = strtok(NULL,delims);                      // parse next value
                                            if (cmd != NULL) {                              // value specified
                                                f[1] = (float)atof(cmd);                    // convert number
                                                cmd = strtok(NULL,delims);                  // parse next value
                                                if (cmd != NULL) {                          // value specified
                                                    f[2] = (float)atof(cmd);                // convert number
                                                    cfg.d = f[2];                           // update D
                                                }
                                                cfg.i = f[1];                               // update I
                                            }
                                            cfg.p = f[0];                                   // update P
                                            WrConfig(&cfg);                                 // save new config/cal info to flash
                                            InitHeater(CASE_HTR,cfg.p,cfg.i,cfg.d);         // update PID info for current heater operation
                                            InitHeater(AUX_HTR,cfg.p,cfg.i,cfg.d);
                                        }
                                        memset(str,0,sizeof(str));          // init string
                                        snprintf(str,sizeof(str)-1,"pid = %1.4g, %1.4g, %1.4g",cfg.p,cfg.i,cfg.d);
                                        SIOtxMsg(str);                      // display new cal values
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                    case LED:                               // show/set LED mode and intensities
                                        cmd = strtok(NULL,delims);                          // parse next value
                                        if (cmd != NULL) {                                  // value specified
                                            uc = (uchar)atoi(cmd);                          // convert number (mode)
                                            cmd = strtok(NULL,delims);                      // parse next value
                                            if (cmd != NULL) {                              // value specified
                                                f[0] = (float)atof(cmd);                    // convert number
                                                cmd = strtok(NULL,delims);                  // parse next value
                                                if (cmd != NULL) {                          // value specified
                                                    f[1] = (float)atof(cmd);                // convert number
                                                    cmd = strtok(NULL,delims);              // parse next value
                                                    if (cmd != NULL) {                      // value specified
                                                        f[2] = (float)atof(cmd);            // convert number
                                                        cfg.blue = f[2];                    // update blue intensity value
                                                    }
                                                    cfg.grn = f[1];                         // update green intensity value
                                                }
                                                cfg.red = f[0];                             // update red intensity value
                                            }
                                            if (uc) {                       // update LED operating mode
                                                cfg.led_mode = FIXED;       // use fixed settings
                                            } else {
                                                cfg.led_mode = AUTO;        // use auto settings
                                            }
                                            WrConfig(&cfg);                 // save new config/cal info to flash
                                        }
                                        memset(str,0,sizeof(str));          // init string
                                        snprintf(str,sizeof(str)-1,"led = %d, %1.1f, %1.1f, %1.1f",cfg.led_mode,cfg.red,cfg.grn,cfg.blue);
                                        SIOtxMsg(str);                      // display new cal values
                                        processed = true;
                                        msg[0] = '\0';                      // clear return message
                                        break;
                                }
                            }
                        }
                        break;
                }
            }
            if (!processed) {                                   // message match not found
                SIOtxMsg("?");
                msg[0] = '\0';                                  // clear return message
            }
        } else {                                                // queue is empty
            SIOrxQ.msg_avail = 0;                               // reset msg counter
        }
    }
    return strlen(msg);                                         // return message length (0 = no msg)
}


// clear (empty) the receive queue
void SIOclrRxQ (void) {
    SIOdisableRxInt();                      // disable receive irq
    SIOrxQ.first = 0;                       // reset queue
    SIOrxQ.last = 0;
    SIOrxQ.msg_avail = 0;
    SIOenableRxInt();                       // enable receive irq
}


// SIO receive interrupt handler
void SIOrxInt (void) __interrupt[USCIAB0RX_VECTOR] {
    uint    last;

    if (UCA0RXBUF == BS) {                                              // destructive backspace
        if (SIOrxQ.first != SIOrxQ.last) {                              // char present to delete
            if (SIOrxQ.last == 0) {                                     // move pointer to remove char
                SIOrxQ.last = MSG_RX_QUEUE_SIZE - 1;
            } else {
                SIOrxQ.last = SIOrxQ.last - 1;
            }
        }
    } else {
        if (UCA0RXBUF != LF) {                                          // keep all chars except LF
            SIOrxQ.buffer[SIOrxQ.last] = UCA0RXBUF;                         // put new char in queue
            if (SIOrxQ.buffer[SIOrxQ.last] == CR) {                         // update message count
                SIOrxQ.msg_avail++;
            }
            SIOrxQ.last = (SIOrxQ.last + 1) % MSG_RX_QUEUE_SIZE;    // update pointer
        }
    }
}


// SIO transmit interrupt handler
void SIOtxInt (void) __interrupt[USCIAB0TX_VECTOR] {
    uint    last;

    if (SIOtxQ.first != SIOtxQ.last) {                                  // buffer not empty, send next char
        UCA0TXBUF = SIOtxQ.buffer[SIOtxQ.first];                        // copy char to TX reg
        SIOtxQ.first = (SIOtxQ.first + 1) % MSG_TX_QUEUE_SIZE;          // update index pointer
    } else {
        SIOdisableTxInt();                                              // buffer empty, disable interrupt
    }
}


// display top-most menu
void DispMenu (void) {

    pauseCnt = MENU_PAUSE;                  // pause unpolled data when displaying menu(s)
    SIOtxMsg("");
    switch (currentMenu) {
        case MAIN_MENU:
            SIOtxStr("NOAA/ESRL/GMD C3WV3 - "); SIOtxMsg((char*)&VERSION);
            SIOtxMsg("[?] = Show menu");
            SIOtxMsg("[cfg] = Configuration Menu");
            SIOtxMsg("[cal] = Calibration Menu");
            SIOtxMsg("[stop] = Stop active filter (start change)");
            SIOtxMsg("[go] = Filter change complete (=n optional)");
            SIOtxMsg("[spot] = Activate specified filter spot");
            SIOtxMsg("[hide] = Stop unpolled data");
            SIOtxMsg("[show] = Enable unpolled data");
            SIOtxMsg("");
            break;
        case CFG_MENU:
            SIOtxMsg("Configuration Menu");
            SIOtxMsg("[?] = Show menu");
            SIOtxMsg("[oint] = Output interval (seconds)");
            SIOtxMsg("[sn] = Serial number");
            SIOtxMsg("[fw] = Firmware version");
            SIOtxMsg("[lpf] = Low-pass filter");
            SIOtxMsg("[hsp] = Case heater setpt (C)");
            SIOtxMsg("[stbl] = Case heater stable window (+/- C)");
            SIOtxMsg("[rst] = reset to defaults");
            SIOtxMsg("[hide] = Stop unpolled data");
            SIOtxMsg("[show] = Enable unpolled data");
            SIOtxMsg("[main] = Return to main menu");
            SIOtxMsg("");
            break;
        case CAL_MENU:
            SIOtxMsg("Calibration Menu");
            SIOtxMsg("[?] = Show menu");
            SIOtxMsg("[Tc] = Case temperature calibration");
            SIOtxMsg("[Ta] = Air temperature calibration");
            SIOtxMsg("[flow] = Flow calibration");
            SIOtxMsg("[pid] = PID gain settings");
            SIOtxMsg("[led] = LED mode and intensity setting");
            SIOtxMsg("[hide] = Stop unpolled data");
            SIOtxMsg("[show] = Enable unpolled data");
            SIOtxMsg("[main] = Return to main menu");
            SIOtxMsg("");
            break;
    }
}
